# Generated by Django 2.1.9 on 2019-06-19 12:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0027_eventrecord_first_saved_at")]

    operations = [
        migrations.AddField(
            model_name="provider", name="operator", field=models.BooleanField(null=True)
        )
    ]



from django.conf.urls import include
from django.urls import path
from rest_framework import routers

from . import authent, provider_api, providers, service_areas, vehicles, event_records


def get_prv_router():
    """Generates a fresh router.

    Enables to register new routes even after .urls has
    been called on the router.
    """
    prv_router = routers.DefaultRouter()
    prv_router.trailing_slash = "/?"  # Make trailing slash optional
    prv_router.register(r"providers", providers.ProviderViewSet)
    prv_router.register(r"service_areas", service_areas.AreaViewSet)
    prv_router.register(r"polygons", service_areas.PolygonViewSet)
    prv_router.register(r"vehicles", vehicles.DeviceViewSet, basename="device")
    prv_router.register(
        r"provider_api", provider_api.ProviderApiViewSet, basename="provider_api"
    )
    prv_router.register(
        r"event_records", event_records.EventRecordViewSet, basename="event_records"
    )
    return prv_router


app_name = "mds_prv_api"

auth_urls = [
    path(
        "long_lived_token/",
        authent.LongLivedTokenView.as_view(),
        name="long_lived_token",
    ),
    path("create_application/", authent.AppCreationView.as_view(), name="create_app"),
    path("revoke_application/", authent.AppCreationView.as_view(), name="revoke_app"),
    path("delete_application/", authent.AppCreationView.as_view(), name="delete_app"),
]

urlpatterns = [path("authent/", include(auth_urls))]
urlpatterns += get_prv_router().urls



# Generated by Django 2.2 on 2019-05-17 08:38

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0023_relabel_api_fields")]

    operations = [
        migrations.AlterField(
            model_name="provider",
            name="device_category",
            field=mds.models.UnboundedCharField(
                blank=True,
                choices=[
                    ("bicycle", "Bicycle"),
                    ("scooter", "Scooter"),
                    ("car", "Car"),
                ],
            ),
        )
    ]



import logging

from rest_framework import mixins
from rest_framework import serializers
from rest_framework import status
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from django.contrib.gis.geos import Point
from django.db.utils import IntegrityError

from mds import db_helpers
from mds import enums, models, provider_mapping
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_AGENCY_API
from mds.apis import utils as apis_utils

logger = logging.getLogger(__name__)


class DeviceSerializer(serializers.Serializer):
    """Expose devices as described in the Agency spec."""

    device_id = serializers.UUIDField(
        source="id", help_text="Provided by Operator to uniquely identify a vehicle."
    )
    provider_id = serializers.UUIDField(help_text="Issued by City and tracked.")
    vehicle_id = serializers.CharField(
        source="identification_number",
        help_text="Vehicle Identification Number (vehicle_id) visible on vehicle.",
    )
    type = serializers.ChoiceField(
        source="category",
        choices=enums.choices(enums.DEVICE_CATEGORY),
        help_text="Vehicle type.",
    )
    propulsion = serializers.ListSerializer(
        child=serializers.ChoiceField(choices=enums.choices(enums.DEVICE_PROPULSION)),
        help_text="Array of Propulsion Type.",
    )
    year = serializers.IntegerField(
        source="year_manufactured", help_text="Year Manufactured."
    )
    mfgr = serializers.CharField(
        source="manufacturer", help_text="Vehicle Manufacturer."
    )
    model = serializers.CharField(help_text="Vehicle Model.")
    status = serializers.ChoiceField(
        source="dn_status",
        choices=enums.choices(enums.DEVICE_STATUS),
        help_text="Current vehicle status.",
    )
    prev_event = serializers.ChoiceField(
        source="latest_event.event_type",
        choices=enums.choices(enums.EVENT_TYPE),
        help_text="Last Vehicle Event.",
        allow_null=True,
    )
    updated = apis_utils.UnixTimestampMilliseconds(
        source="latest_event.saved_at",
        help_text="Date of last event update as Unix Timestamp (milliseconds).",
        allow_null=True,
    )


class DeviceRegisterSerializer(serializers.Serializer):
    """Receive a new device to create from a provider."""

    device_id = serializers.UUIDField(
        source="id", help_text="Provided by Operator to uniquely identify a vehicle."
    )
    vehicle_id = serializers.CharField(
        source="identification_number",
        help_text="Vehicle Identification Number (vehicle_id) visible on vehicle.",
    )
    type = serializers.ChoiceField(
        source="category",
        choices=enums.choices(enums.DEVICE_CATEGORY),
        help_text="Vehicle type.",
    )
    propulsion = serializers.ListSerializer(
        child=serializers.ChoiceField(choices=enums.choices(enums.DEVICE_PROPULSION)),
        help_text="Array of Propulsion Type.",
    )
    year = serializers.IntegerField(
        required=False, source="year_manufactured", help_text="Year Manufactured."
    )
    mfgr = serializers.CharField(
        required=False, source="manufacturer", help_text="Vehicle Manufacturer."
    )
    model = serializers.CharField(required=False, help_text="Vehicle Model.")

    def create(self, validated_data):
        provider_id = self.context["request"].user.provider_id
        try:
            return models.Device.objects.create(
                provider_id=provider_id, **validated_data
            )
        except IntegrityError:
            detail = f"A vehicle with id={validated_data['id']} is already registered"
            raise apis_utils.AlreadyRegisteredError({"already_registered": detail})


class GPSSerializer(serializers.Serializer):
    """GPS data inside the telemetry frame sent by the provider."""

    lat = serializers.FloatField()
    lng = serializers.FloatField()
    altitude = serializers.FloatField(required=False, help_text="in meters")
    heading = serializers.FloatField(
        required=False, min_value=0, help_text="degrees, starting at 0 at true North"
    )
    speed = serializers.FloatField(required=False, help_text="in meters/second")
    hdop = serializers.FloatField(
        required=False,
        min_value=1,
        source="accuracy",
        help_text="Horizontal GPS accuracy",
    )
    satellites = serializers.IntegerField(
        required=False, min_value=0, help_text="Number of GPS satellites"
    )

    def validate(self, value):
        attrs = super().validate(value)

        # Some providers may mistake latitude and longitude
        provider = self.context.get("provider")
        if provider and provider.agency_api_configuration.get("swap_lat_lng"):
            attrs["lat"], attrs["lng"] = attrs["lng"], attrs["lat"]

        # Now we can validate (this will not catch valid inversions)
        if attrs["lat"] < -90.0 or attrs["lat"] > 90.0:
            raise ValidationError(
                {"lat": "Latitude is outside [-90 90]: %s" % attrs["lat"]}
            )
        if attrs["lng"] < -180.0 or attrs["lng"] > 180.0:
            raise ValidationError(
                {"lng": "Longitude is outside [-180 180]: %s" % attrs["lng"]}
            )

        return attrs


def gps_to_gis_point(gps_data):
    if gps_data:
        # TODO(lip): maybe use altitude as z ?
        return Point(gps_data["lng"], gps_data["lat"], srid=4326)
    return None


class DeviceTelemetrySerializer(serializers.Serializer):
    """Telemetry frame from event and telemetry endpoints."""

    device_id = serializers.UUIDField()
    gps = GPSSerializer()
    timestamp = apis_utils.UnixTimestampMilliseconds(
        help_text="Unix timestamp in milliseconds"
    )
    charge = serializers.FloatField(
        required=False,
        source="battery_pct",
        min_value=0,
        max_value=1,
        help_text="Percent battery charge of vehicle, expressed between 0 and 1",
    )


class DeviceEventSerializer(serializers.Serializer):
    """Receive a new event pushed by a provider."""

    event_type = serializers.ChoiceField(
        choices=enums.choices(enums.EVENT_TYPE), help_text="Vehicle event."
    )
    event_type_reason = serializers.ChoiceField(
        choices=enums.choices(enums.EVENT_TYPE_REASON),
        help_text="Vehicle event type reason.",
        required=False,
        allow_null=True,
    )
    timestamp = apis_utils.UnixTimestampMilliseconds(
        help_text="Timestamp of the last event update"
    )
    telemetry = DeviceTelemetrySerializer(
        write_only=True, help_text="Single point of telemetry."
    )
    trip_id = serializers.UUIDField(
        required=False,
        allow_null=True,
        help_text=(
            "UUID provided by Operator to uniquely identify the trip. "
            "Required for trip_* events."
        ),
    )

    def get_event(self, validated_data):
        # We first check whether the provider uses the old or the new agency
        # event_type(s) and event_type_reason(s).
        # A provider uses the old version if agency_api_version == "draft" in its
        # api_configuration.
        provider_id = self.context["request"].user.provider_id
        provider = models.Provider.objects.get(id=provider_id)
        old_version = provider.api_configuration.get("agency_api_version") == "draft"
        event_type = validated_data.get("event_type")
        event_type_reason = validated_data.get("event_type_reason")
        if old_version:
            return (event_type, None)
        else:
            event = (event_type, event_type_reason)
            if event not in provider_mapping.AGENCY_EVENT_TO_PROVIDER_REASON.keys():
                # This should be avoided if possible
                msg = f"The event ({event[0]}, {event[1]}) is not in the mapping."
                logger.warning(msg)
                raise ValidationError(msg)
            return event

    def create(self, validated_data):
        event_type, event_type_reason = self.get_event(validated_data)

        device = self.context["device"]
        event_record = models.EventRecord(
            timestamp=validated_data["timestamp"],
            point=gps_to_gis_point(validated_data["telemetry"].get("gps", {})),
            device_id=device.id,
            event_type=event_type,
            event_type_reason=event_type_reason,
            properties={
                "telemetry": validated_data["telemetry"],
                "trip_id": validated_data.get("trip_id"),
            },
        )
        db_helpers.upsert_event_records(
            [event_record], enums.EVENT_SOURCE.agency_api.name, on_conflict_update=True
        )

        # We don't get the created event record but we need to return it
        return models.EventRecord.objects.get(
            device=device, timestamp=validated_data["timestamp"]
        )


class DeviceEventResponseSerializer(serializers.Serializer):
    """Response format for the event endpoint."""

    device_id = serializers.UUIDField()
    status = serializers.ChoiceField(
        source="updated_status", choices=enums.choices(enums.DEVICE_STATUS)
    )


class DeviceTelemetryInputSerializer(serializers.Serializer):
    """Receive a new telemetry pushed by a provider."""

    data = DeviceTelemetrySerializer(many=True)

    def create(self, validated_data):
        provider_id = self.context["request"].user.provider_id
        unknown_ids = [
            str(id)
            for id in (
                set(t["device_id"] for t in validated_data["data"])
                - set(
                    models.Device.objects.filter(
                        id__in=[t["device_id"] for t in validated_data["data"]]
                    )
                    .filter(provider_id=provider_id)
                    .values_list("id", flat=True)
                )
            )
        ]
        if unknown_ids:
            raise ValidationError(
                {"data.device_id": "Unknown ids: %s" % " ".join(unknown_ids)}
            )

        event_records = (
            models.EventRecord(
                timestamp=telemetry["timestamp"],
                point=gps_to_gis_point(telemetry.get("gps", {})),
                device_id=telemetry["device_id"],
                event_type=enums.EVENT_TYPE.telemetry.name,
                properties={"telemetry": telemetry, "trip_id": None},
            )
            for telemetry in validated_data["data"]
        )
        db_helpers.upsert_event_records(
            event_records, enums.EVENT_SOURCE.agency_api.name, on_conflict_update=True
        )

        # We don't have the created event records,
        # but we will return an empty response anyway (cf. DeviceViewSet)
        return []


class DeviceViewSet(
    apis_utils.MultiSerializerViewSetMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    mixins.CreateModelMixin,
    viewsets.GenericViewSet,
):

    queryset = models.Device.objects.with_latest_events()
    permission_classes = (require_scopes(SCOPE_AGENCY_API),)
    lookup_field = "id"
    serializer_class = DeviceSerializer
    serializers_mapping = {
        "list": {"response": DeviceSerializer},
        "retrieve": {"response": DeviceSerializer},
        "create": {
            "request": DeviceRegisterSerializer,
            "response": apis_utils.EmptyResponseSerializer,
        },
        "event": {
            "request": DeviceEventSerializer,
            "response": DeviceEventResponseSerializer,
        },
        "telemetry": {
            "request": DeviceTelemetryInputSerializer,
            "response": apis_utils.EmptyResponseSerializer,
        },
    }

    def create(self, *args, **kwargs):
        return self._create(*args, **kwargs)

    @action(detail=True, methods=["post", "options"])
    def event(self, request, id):
        """Endpoint to receive an event from a provider."""
        provider_id = request.user.provider_id
        device = models.Device.objects.filter(provider_id=provider_id, id=id).last()
        if not device:
            return Response(data={}, status=404)

        provider = models.Provider.objects.get(pk=provider_id)
        request_serializer = self.get_serializer(
            data=request.data,
            context={
                "device": device,
                "request_or_response": "request",
                "provider": provider,
            },
        )
        request_serializer.is_valid(raise_exception=True)
        instance = request_serializer.save()
        response_serializer = self.get_serializer(
            instance=instance, context={"request_or_response": "response"}
        )
        return Response(response_serializer.data, status=status.HTTP_201_CREATED)

    @action(detail=False, methods=["post", "options"])
    def telemetry(self, request):
        """Endpoint to receive a telemetry from a provider."""
        context = self.get_serializer_context()  # adds the request to the context
        context["request_or_response"] = "request"
        provider_id = request.user.provider_id
        context["provider"] = models.Provider.objects.get(pk=provider_id)
        serializer = self.get_serializer(data=request.data, context=context)
        serializer.is_valid(raise_exception=True)
        instance = serializer.save()
        response_serializer = self.get_serializer(
            instance=instance, context={"request_or_response": "response"}
        )
        return Response(response_serializer.data, status=status.HTTP_201_CREATED)

    def get_queryset(self):
        queryset = super().get_queryset()
        provider_id = getattr(self.request.user, "provider_id", None)
        if provider_id:
            queryset = queryset.filter(provider_id=provider_id)
        else:
            queryset = queryset.none()

        return queryset



"""
Everything related to support older Provider API versions
by translating them to the latest supported version.
"""
from semantic_version import Version


def translate_data(data, version):
    version = Version(version)

    if version >= Version("0.3.0") and version < Version("0.4.0"):
        return data

    if version >= Version("0.2.0"):
        # The only two noticeable changes from our point of view are:
        # - timestamps converted from floating-point seconds to milliseconds;
        # - "trip_ids" now is a single "trip_id"
        for status_change in data["status_changes"]:
            # We were already expecting milliseconds in the 0.2 implementation
            if "." in str(status_change["event_time"]):
                status_change["event_time"] = round(status_change["event_time"] * 1000)
            # Keep only the first (and probably only) trip ID
            associated_trip = status_change.pop("associated_trips", None)
            if associated_trip:
                status_change["associated_trip"] = associated_trip[0]
        return data

    raise NotImplementedError(version)



from collections import namedtuple
import datetime
import uuid

from django.utils import timezone
import jwt

from . import models
from . import settings


AuthMean = namedtuple("AuthMean", ("key", "algorithm"))


def generate_jwt_with_payload(application, token_duration, user=None):
    payload = _generate_payload(application, token_duration, user=user)
    return _generate_jwt(payload), payload


def generate_jwt(application, token_duration, save=False):
    token, payload = generate_jwt_with_payload(application, token_duration)
    if save:
        models.AccessToken.objects.create(
            token=token,
            jti=payload["jti"],
            application=application,
            expires=datetime.datetime.fromtimestamp(
                payload["exp"], tz=timezone.now().tzinfo
            ),
        )
    return token


def _generate_jwt(payload):
    auth_mean = _get_auth_mean()
    # returns a string because it needs to be JSON serializable by oauthlib
    return jwt.encode(payload, auth_mean.key, algorithm=auth_mean.algorithm).decode(
        "utf-8"
    )


def _get_auth_mean():
    if settings.AUTHENT_RSA_PRIVATE_KEY:
        return AuthMean(settings.AUTHENT_RSA_PRIVATE_KEY, "RS256")
    return AuthMean(settings.AUTHENT_SECRET_KEY, "HS256")


def _generate_payload(application, token_duration, user):
    payload = {"jti": str(uuid.uuid4()), "scope": application.scopes_string}
    if user:
        payload["sub"] = "user:%s" % user.username
        payload["name"] = user.get_full_name() or user.username
    else:
        payload["sub"] = ("application:%s" % application.client_id,)
    if application.owner:
        payload["app_owner"] = str(application.owner)
    if token_duration:
        expiry = timezone.now() + token_duration
        payload["exp"] = int(expiry.timestamp())

    return payload



from django_filters import rest_framework as filters
from drf_yasg.utils import swagger_serializer_method
from rest_framework import serializers, viewsets

from mds import enums, models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_PRV_API
from mds.apis import utils

from typing import List, Dict


class DeviceFilter(filters.FilterSet):
    id = filters.CharFilter(lookup_expr="icontains")
    category = utils.ChoicesInFilter(choices=enums.choices(enums.DEVICE_CATEGORY))
    provider = utils.UUIDInFilter()
    status = utils.ChoicesInFilter(
        "dn_status", choices=enums.choices(enums.DEVICE_STATUS)
    )
    registrationDateFrom = filters.IsoDateTimeFilter(
        "registration_date", lookup_expr="gte"
    )
    registrationDateTo = filters.IsoDateTimeFilter(
        "registration_date", lookup_expr="lte"
    )

    class Meta:
        model = models.Device
        fields = [
            "id",
            "category",
            "provider",
            "status",
            "registrationDateFrom",
            "registrationDateTo",
        ]


class DeviceSerializer(serializers.ModelSerializer):
    id = serializers.UUIDField(help_text="Unique device identifier (UUID)")
    model = serializers.CharField(required=False, help_text="Vehicle model")
    identification_number = serializers.CharField(
        help_text="VIN (Vehicle Identification Number)"
    )
    category = serializers.ChoiceField(
        enums.choices(enums.DEVICE_CATEGORY), help_text="Device type"
    )
    propulsion = serializers.ListField(
        child=serializers.ChoiceField(enums.choices(enums.DEVICE_PROPULSION)),
        help_text="Propulsion type(s)",
    )
    provider_id = serializers.UUIDField(
        source="provider.id", help_text="ID of the service provider of the device"
    )
    provider_name = serializers.CharField(
        source="provider.name", help_text="Name of the service provider of the device"
    )
    registration_date = serializers.DateTimeField(help_text="Device registration date")
    last_telemetry_date = serializers.DateTimeField(
        source="dn_gps_timestamp", help_text="Latest GPS timestamp", allow_null=True
    )
    position = utils.PointSerializer(
        source="dn_gps_point", help_text="Latest GPS position"
    )
    status = serializers.ChoiceField(
        enums.choices(enums.DEVICE_STATUS),
        source="dn_status",
        help_text="Latest status",
        allow_null=True,
    )
    battery = serializers.FloatField(
        source="dn_battery_pct", help_text="Percentage between 0 and 1", allow_null=True
    )

    class Meta:
        model = models.Device
        fields = (
            "id",
            "provider_id",
            "provider_name",
            "model",
            "identification_number",
            "category",
            "propulsion",
            "status",
            "position",
            "registration_date",
            "last_telemetry_date",
            "battery",
        )


class RetrieveDeviceSerializer(DeviceSerializer):
    areas = serializers.SerializerMethodField()
    provider_logo = serializers.CharField(
        source="provider.logo_b64",
        help_text="logo in base 64 of the service provider of the device",
    )

    # swagger_serializer_method needed because drf-yasg type hinting detection
    # is broken in 3.7
    @swagger_serializer_method(
        serializer_or_field=serializers.ListField(
            child=serializers.DictField(child=serializers.CharField())
        )
    )
    def get_areas(self, obj) -> List[Dict[str, str]]:
        if not obj.dn_gps_point:
            return []
        areas = (
            models.Area.objects.filter(polygons__geom__contains=obj.dn_gps_point)
            .order_by("label", "id")
            .distinct("label", "id")
        )
        return list(areas.values("id", "label"))

    class Meta:
        model = models.Device
        fields = DeviceSerializer.Meta.fields + ("areas", "provider_logo")


class DeviceViewSet(utils.MultiSerializerViewSetMixin, viewsets.ReadOnlyModelViewSet):
    permission_classes = (require_scopes(SCOPE_PRV_API),)
    lookup_field = "id"
    serializer_class = DeviceSerializer
    pagination_class = utils.LimitOffsetPagination
    filter_backends = (filters.DjangoFilterBackend,)

    filterset_class = DeviceFilter
    queryset = models.Device.objects.select_related("provider").all()
    serializers_mapping = {
        "list": {"response": DeviceSerializer},
        "retrieve": {"response": RetrieveDeviceSerializer},
    }



import django_filters.rest_framework
from rest_framework import serializers
from rest_framework import viewsets
from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_PRV_API


class ProviderColorsSerializer(serializers.Serializer):
    primary = serializers.CharField(required=False)
    secondary = serializers.CharField(required=False)


class ProviderSerializer(serializers.ModelSerializer):

    id = serializers.UUIDField(
        required=True, help_text="Unique provider identifier (UUID)"
    )
    name = serializers.CharField(help_text="Name of the Provider")
    logo_b64 = serializers.CharField(
        required=False, help_text="Logo of provider base64 encoded"
    )
    operator = serializers.BooleanField(
        required=False, help_text="This provider is a device operator"
    )
    colors = ProviderColorsSerializer(
        default=dict, help_text="colors for distinguishing providers on map"
    )
    device_categories = serializers.DictField(serializers.IntegerField())

    def create(self, validated_data):
        instance = self.Meta.model(**validated_data)
        instance.save()
        return instance

    class Meta:
        model = models.Provider
        fields = (
            "id",
            "name",
            "colors",
            "logo_b64",
            "base_api_url",
            "api_authentication",
            "api_configuration",
            "agency_api_authentication",
            "device_categories",
            "operator",
        )


class ProviderViewSet(viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_PRV_API),)
    queryset = models.Provider.objects.with_device_categories()
    lookup_field = "id"
    serializer_class = ProviderSerializer
    filter_backends = (django_filters.rest_framework.DjangoFilterBackend,)
    filterset_fields = ("operator",)



# Generated by Django 2.1.5 on 2019-03-15 15:52

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0013_provider_last_start_time_polled'),
    ]

    operations = [
        migrations.AlterField(
            model_name='area',
            name='label',
            field=mds.models.UnboundedCharField(blank=True, db_index=True, default=''),
        ),
        migrations.AlterField(
            model_name='device',
            name='manufacturer',
            field=mds.models.UnboundedCharField(default=''),
        ),
        migrations.AlterField(
            model_name='device',
            name='model',
            field=mds.models.UnboundedCharField(default=''),
        ),
        migrations.AlterField(
            model_name='eventrecord',
            name='event_type',
            field=mds.models.UnboundedCharField(choices=[('service_start', 'Service start'), ('trip_end', 'Trip end'), ('rebalance_drop_off', 'Rebalance drop off'), ('maintenance_drop_off', 'Maintenance drop off'), ('cancel_reservation', 'Cancel reservation'), ('reserve', 'Reserve'), ('trip_start', 'Trip start'), ('trip_enter', 'Trip enter'), ('trip_leave', 'Trip leave'), ('register', 'Register'), ('low_battery', 'Low battery'), ('maintenance', 'Maintenance'), ('service_end', 'Service end'), ('rebalance_pick_up', 'Rebalance pick up'), ('maintenance_pick_up', 'Maintenance pick up'), ('deregister', 'Deregister'), ('telemetry', 'Received telemetry'), ('battery_charged', 'Battery charged')]),
        ),
        migrations.AlterField(
            model_name='polygon',
            name='label',
            field=mds.models.UnboundedCharField(blank=True, db_index=True, default=''),
        ),
        migrations.AlterField(
            model_name='provider',
            name='name',
            field=mds.models.UnboundedCharField(default=''),
        ),
    ]



# Generated by Django 2.1.5 on 2019-03-20 15:47

import django.contrib.postgres.fields.jsonb
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0015_null_dn_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='provider',
            name='agency_api_configuration',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=mds.models.agency_api_configuration_default, verbose_name='Agency API Configuration'),
        ),
        migrations.AlterField(
            model_name='provider',
            name='agency_api_authentication',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=mds.models.agency_api_authentication_default, verbose_name='Agency API Authentication'),
        ),
        migrations.AlterField(
            model_name='provider',
            name='api_authentication',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=mds.models.provider_api_authentication_default, verbose_name='Provider API Authentication'),
        ),
        migrations.AlterField(
            model_name='provider',
            name='api_configuration',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=mds.models.provider_api_configuration_default, verbose_name='Provider API Configuration'),
        ),
        migrations.AlterField(
            model_name='provider',
            name='base_api_url',
            field=mds.models.UnboundedCharField(default='', verbose_name='Base Provider API URL'),
        ),
    ]



from django.contrib import admin
from . import models

from uuid import UUID


def is_uuid(uuid_string, version=4):
    try:
        UUID(uuid_string, version=version)
    except ValueError:
        return False
    return True


@admin.register(models.Provider)
class ProviderAdmin(admin.ModelAdmin):
    list_display = ["id", "name", "operator"]
    list_filter = ["operator"]
    ordering = ["name"]


@admin.register(models.Device)
class DeviceAdmin(admin.ModelAdmin):
    list_display = ["id", "provider_name", "identification_number", "category"]
    list_filter = ["provider", "category"]
    search_fields = ["id", "identification_number"]
    list_select_related = ["provider"]

    def provider_name(self, obj):
        return obj.provider.name

    provider_name.short_description = "Provider"


@admin.register(models.EventRecord)
class EventRecordAdmin(admin.ModelAdmin):
    list_display = [
        "saved_at",
        "timestamp",
        "provider",
        "device_id",
        "event_type",
        "event_type_reason",
    ]
    list_filter = ["device__provider", "event_type", "event_type_reason"]
    list_select_related = ["device__provider"]
    search_fields = ["device__id", "device__identification_number"]

    def get_search_results(self, request, queryset, search_term):
        if not search_term:
            return super().get_search_results(request, queryset, search_term)
        custom_queryset = get_devices_queryset_search_results(self, search_term)
        return super().get_search_results(request, custom_queryset, search_term)

    def provider(self, obj):
        return obj.device.provider.name

    def device_id(self, obj):
        return obj.device.id


# to use when searching for devices in get_search_results
# as a relationship to self.model
def get_devices_queryset_search_results(self, search_term):
    custom_queryset = self.model.objects.select_related(*self.list_select_related)
    if is_uuid(search_term):
        return custom_queryset.filter(device_id=search_term)
    else:
        return custom_queryset.filter(device__identification_number=search_term)


@admin.register(models.Area)
class AreaAdmin(admin.ModelAdmin):
    list_display = ["id", "label"]
    ordering = ["label"]


@admin.register(models.Polygon)
class PolygonAdmin(admin.ModelAdmin):
    list_display = ["id", "label"]
    ordering = ["label"]



# Generated by Django 2.1.7 on 2019-04-11 08:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0018_allow_blank_provider_base_api_url'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='area',
            name='creation_date',
        ),
        migrations.RemoveField(
            model_name='area',
            name='deletion_date',
        ),
        migrations.RemoveField(
            model_name='polygon',
            name='creation_date',
        ),
        migrations.RemoveField(
            model_name='polygon',
            name='deletion_date',
        ),
    ]



# Generated by Django 2.1.10 on 2019-07-23 17:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0034_post_default_device_saved_at")]

    operations = [
        migrations.AddField(
            model_name="provider",
            name="last_event_time_polled",
            field=models.DateTimeField(
                blank=True,
                null=True,
                verbose_name="Last event time polled (start_time)",
            ),
        ),
        migrations.AddField(
            model_name="provider",
            name="last_recorded_polled",
            field=models.DateTimeField(
                blank=True,
                null=True,
                verbose_name="Last recorded polled (start_recorded)",
            ),
        ),
        migrations.AddField(
            model_name="provider",
            name="last_skip_polled",
            field=models.PositiveIntegerField(
                blank=True, null=True, verbose_name="Last skip polled (total_events)"
            ),
        ),
    ]



"""
Classes representing authentication means.

For now only JWT based authentication is supported as required by MDS specification:
https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/agency#authorization

When using JWT as access tokens, the common usage is to not check with the
authorization server if the JWT is revoked and use short-lived tokens instead.
Because we fear most tokens in real world use of MDS will be long-lived  (contrary
to the spirit of JWT tokens), we optionally accept an introspection endpoint that will
be called to check token validity in cas it has been revoked

We could potentially handle access tokens that are not JWT.
In this case, the introspect_url would be required to retrieve the user scopes
"""
from typing import Optional


class BaseAuthMean:
    introspect_url: Optional[str]


class SecretKeyJwtBaseAuthMean(BaseAuthMean):
    secret_key: str

    def __init__(self, secret_key: str, introspect_url: str = None):
        self.secret_key = secret_key
        self.introspect_url = introspect_url


class PublicKeyJwtBaseAuthMean(BaseAuthMean):
    public_key: str

    def __init__(self, public_key: str, introspect_url: str = None):
        self.public_key = public_key
        self.introspect_url = introspect_url



# Generated by Django 2.2b1 on 2019-03-01 13:16

import django.contrib.postgres.fields
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0011_added_index_to_polygon_and_area_models'),
    ]

    operations = [
        migrations.AddField(
            model_name='area',
            name='area_type',
            field=mds.models.UnboundedCharField(choices=[('unrestricted', 'Unrestricted'), ('restricted', 'Restricted'), ('preferred_pick_up', 'Preferred pick up'), ('preferred_drop_off', 'Preferred drop off')], default='unrestricted'),
        ),
        migrations.AlterField(
            model_name='device',
            name='dn_status',
            field=mds.models.UnboundedCharField(choices=[('available', 'Available'), ('reserved', 'Reserved'), ('unavailable', 'Unavailable'), ('removed', 'Removed'), ('trip', 'Trip'), ('elsewhere', 'Elsewhere'), ('inactive', 'Inactive'), ('unknown', 'Unknown')], default='unavailable'),
        ),
        migrations.AlterField(
            model_name='device',
            name='propulsion',
            field=django.contrib.postgres.fields.ArrayField(base_field=mds.models.UnboundedCharField(choices=[('human', 'Human'), ('electric_assist', 'Electric Assist'), ('electric', 'Electric'), ('combustion', 'Combustion')]), size=None),
        ),
        migrations.AlterField(
            model_name='eventrecord',
            name='event_type',
            field=mds.models.UnboundedCharField(choices=[('service_start', 'Service start'), ('trip_end', 'Trip end'), ('rebalance_drop_off', 'Rebalance drop off'), ('maintenance_drop_off', 'Maintenance drop off'), ('cancel_reservation', 'Cancel reservation'), ('reserve', 'Reserve'), ('trip_start', 'Trip start'), ('trip_enter', 'Trip enter'), ('trip_leave', 'Trip leave'), ('register', 'Register'), ('low_battery', 'Low battery'), ('maintenance', 'Maintenance'), ('service_end', 'Service end'), ('rebalance_pick_up', 'Rebalance pick up'), ('maintenance_pick_up', 'Maintenance pick up'), ('deregister', 'Deregister'), ('telemetry', 'Received telemetry'), ('battery_ok', 'Battery OK')]),
        ),
    ]



# Generated by Django 2.1.8 on 2019-05-28 14:52

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0025_remove_provider_device_category'),
    ]

    operations = [
        migrations.AlterField(
            model_name='polygon',
            name='geom',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(srid=4326),
        ),
    ]



# Generated by Django 2.1.4 on 2019-01-24 09:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("authent", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="accesstoken",
            name="revoked_after",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="application",
            name="owner",
            field=models.UUIDField(
                blank=True,
                db_index=True,
                help_text="Unique identifier for the owner of the application, not necessarily a django User.",
                null=True,
            ),
        ),
    ]



import random

from rest_framework import serializers
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from django.db import IntegrityError

from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_PRV_API
from mds.apis import utils

from rest_framework import filters


class PolygonRequestSerializer(serializers.ModelSerializer):
    """What we expect for a geographic polygon.
    """

    label = serializers.CharField(help_text="Name of the polygon")
    geom = utils.PolygonSerializer(help_text="GeoJSON Polygon")
    areas = serializers.PrimaryKeyRelatedField(many=True, queryset=models.Area.objects)

    class Meta:
        fields = ("geom", "label", "areas")
        model = models.Polygon

    def create(self, validated_data):
        instance = self.Meta.model(
            label=validated_data["label"],
            geom=utils.get_geos_multipolygon(str(validated_data["geom"])),
        )
        instance.save()
        return instance

    def update(self, instance, validated_data):
        if validated_data.get("label"):
            instance.label = validated_data["label"]
        if validated_data.get("geom"):
            instance.geom = utils.get_geos_multipolygon(str(validated_data["geom"]))
        if validated_data.get("areas"):
            areas = validated_data.pop("areas", [])
            instance.areas.set(areas)
        instance.save()
        return instance


class PolygonResponseSerializer(serializers.Serializer):
    """A representation of a geographic polygon.
    """

    id = serializers.UUIDField(help_text="Unique Polygon identifier (UUID)")
    label = serializers.CharField(help_text="Name of the polygon")
    geom = utils.PolygonSerializer(help_text="GeoJSON Polygon")
    areas = serializers.PrimaryKeyRelatedField(many=True, read_only=True)
    properties = serializers.DictField()

    class Meta:
        fields = ("id", "label", "geom", "areas", "properties")


class PolygonsImportRequestSerializer(serializers.Serializer):

    polygons = PolygonRequestSerializer(many=True)

    class Meta:
        fields = "polygons"


class PolygonViewSet(utils.MultiSerializerViewSetMixin, viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_PRV_API),)
    queryset = models.Polygon.objects.prefetch_related("areas").all()
    filter_backends = (filters.OrderingFilter,)
    ordering_fields = ("label",)
    ordering = "label"
    lookup_field = "id"
    serializer_class = PolygonResponseSerializer
    serializers_mapping = {
        "list": {"response": PolygonResponseSerializer},
        "retrieve": {"response": PolygonResponseSerializer},
        "create": {
            "request": PolygonRequestSerializer,
            "response": utils.EmptyResponseSerializer,
        },
        "update": {
            "request": PolygonRequestSerializer,
            "response": utils.EmptyResponseSerializer,
        },
        "import_polygons": {
            "request": PolygonsImportRequestSerializer,
            "response": utils.EmptyResponseSerializer,
        },
    }

    def create(self, *args, **kwargs):
        return super()._create(*args, **kwargs)

    def update(self, *args, **kwargs):
        return super()._update(*args, **kwargs)

    @action(methods=["post"], url_path="import", detail=False)
    def import_polygons(self, request, pk=None):
        polygons = request.data.get("polygons", None)

        if not isinstance(polygons, list):
            return Response(status=400)

        try:
            polygons_to_create = []
            for polygon in polygons:
                geom = polygon.get("geom", None)
                if geom and geom["type"] in ["Polygon", "MultiPolygon"]:
                    areas = []
                    for area_label in polygon.get("areas", []):
                        defaults = {"color": "#%06x" % random.randint(0, 0xFFFFFF)}
                        # Create new Area if doesn't exist (based on label)
                        area = models.Area.objects.get_or_create(
                            label=area_label, defaults=defaults
                        )[0]
                        areas.append(area)
                    poly = models.Polygon(
                        label=polygon.get("label", ""),
                        properties=polygon.get("properties", {}),
                        geom=utils.get_geos_multipolygon(str(geom)),
                    )
                    poly.areas.set([a.id for a in areas])
                    polygons_to_create.append(poly)
            models.Polygon.objects.bulk_create(polygons_to_create)
        except IntegrityError as ex:
            return Response(exception=ex, status=500)

        return Response({"message": "ok"})


class AreaRequestSerializer(serializers.ModelSerializer):
    """A service area, composed of a group of Polygons.
    """

    label = serializers.CharField(help_text="Name of the Area")
    polygons = serializers.PrimaryKeyRelatedField(
        many=True, queryset=models.Polygon.objects
    )
    color = serializers.CharField(required=False, help_text="Color of the Area")

    class Meta:
        fields = ("label", "polygons", "color")
        model = models.Area

    def create(self, validated_data):
        instance = self.Meta.model(
            label=validated_data["label"], color=validated_data["color"]
        )
        instance.save()
        polygons = validated_data.get("polygons", [])
        instance.polygons.set(polygons)
        instance.save()
        return instance

    def update(self, instance, validated_data):
        if validated_data.get("label"):
            instance.label = validated_data["label"]
        if validated_data.get("color"):
            instance.color = validated_data["color"]
        if "polygons" in validated_data:
            polygons = validated_data.get("polygons", [])
            instance.polygons.set(polygons)
        instance.save()
        return instance


class AreaResponseSerializer(serializers.Serializer):
    """A service area, composed of a group of Polygons.
    """

    id = serializers.UUIDField(help_text="Unique Area identifier (UUID)")
    label = serializers.CharField(help_text="Name of the Area")
    polygons = PolygonResponseSerializer(many=True)
    color = serializers.CharField(help_text="Color of the Area")

    class Meta:
        fields = ("id", "label", "polygons", "color")


class AreaViewSet(utils.MultiSerializerViewSetMixin, viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_PRV_API),)
    queryset = models.Area.objects.prefetch_related("polygons").all()
    filter_backends = (filters.OrderingFilter,)
    ordering_fields = ("label",)
    ordering = "label"
    lookup_field = "id"
    serializer_class = AreaResponseSerializer
    serializers_mapping = {
        "list": {"response": AreaResponseSerializer},
        "retrieve": {"response": AreaResponseSerializer},
        "create": {
            "request": AreaRequestSerializer,
            "response": utils.EmptyResponseSerializer,
        },
        "update": {
            "request": AreaRequestSerializer,
            "response": utils.EmptyResponseSerializer,
        },
    }

    def create(self, *args, **kwargs):
        return super()._create(*args, **kwargs)

    def update(self, *args, **kwargs):
        return super()._update(*args, **kwargs)



# Generated by Django 2.1.5 on 2019-03-14 15:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0012_update_to_mds_0_3_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='provider',
            name='last_start_time_polled',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.1.3 on 2019-06-18 13:52

import django.contrib.postgres.functions
from django.db import migrations, models


class Migration(migrations.Migration):
    # concurrent index creation cannot be made inside a transaction
    atomic = False

    dependencies = [("mds", "0030_post_provider_operator_default")]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterField(
                    model_name="eventrecord",
                    name="saved_at",
                    field=models.DateTimeField(
                        db_index=True,
                        default=django.contrib.postgres.functions.TransactionNow,
                    ),
                )
            ],
            database_operations=[
                migrations.RunSQL(
                    sql="""
                    CREATE INDEX CONCURRENTLY "mds_eventrecord_saved_at_f421e5fd"
                    ON "mds_eventrecord" ("saved_at");
                    """
                )
            ],
        )
    ]



# Generated by Django 2.1.11 on 2019-08-16 14:14

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0037_remove_provider_last_start_time_polled")]

    operations = [
        migrations.AddField(
            model_name="eventrecord",
            name="event_type_reason",
            field=mds.models.UnboundedCharField(
                blank=True,
                choices=[
                    ("low_battery", "Low battery"),
                    ("maintenance", "Maintenance"),
                    ("compliance", "Compliance"),
                    ("off_hours", "Off hours"),
                    ("rebalance", "Rebalance"),
                    ("charge", "Charge"),
                    ("missing", "Missing"),
                    ("decommissioned", "Decommissioned"),
                ],
                null=True,
            ),
        )
    ]



# Generated by Django 2.2 on 2019-07-03 11:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0033_post_fill_device_saved_at")]

    operations = [
        migrations.AlterField(
            model_name="device",
            name="saved_at",
            field=models.DateTimeField(db_index=True, auto_now=True),
        )
    ]



# Generated by Django 2.2 on 2019-07-03 09:36

from django.contrib.postgres.functions import TransactionNow
from django.db import migrations


def fill_device_saved_at(apps, schema_editor):
    Device = apps.get_model("mds", "Device")

    # We don't have a sequence field to split the table in batches
    start_id = "0" * 32
    for end_id in "123456789abcdef":
        end_id *= 32
        qs = Device.objects.filter(id__gte=start_id, id__lt=end_id)
        qs.update(saved_at=TransactionNow())
        start_id = end_id
    # Then in theory "ffffffff..." itself


class Migration(migrations.Migration):
    atomic = False

    dependencies = [("mds", "0032_pre_device_saved_at")]

    operations = [migrations.RunPython(fill_device_saved_at, migrations.RunPython.noop)]



from mds.enums import (
    DEVICE_STATUS,
    EVENT_TYPE,
    EVENT_TYPE_REASON,
    PROVIDER_EVENT_TYPE_REASON,
)


# Mappings between Provider and Agency
#
# When we poll providers (with provider API) we cast the event reason
# with PROVIDER_REASON_TO_AGENCY_EVENT because enums don't match
# and we use Agency API nomenclature
# Then when we are ourselves polled we use AGENCY_EVENT_TO_PROVIDER_REASON
# and PROVIDER_EVENT_TYPE_REASON_TO_EVENT_TYPE to be compliant
# with Provider API nomenclature

# Converts a Provider event_type_reason to an Agency event_type
# and event_type_reason (when needed), order by key as in:
# https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/provider
PROVIDER_REASON_TO_AGENCY_EVENT = {
    # available #
    PROVIDER_EVENT_TYPE_REASON.service_start.name: (EVENT_TYPE.service_start.name,),
    PROVIDER_EVENT_TYPE_REASON.user_drop_off.name: (EVENT_TYPE.trip_end.name,),
    PROVIDER_EVENT_TYPE_REASON.rebalance_drop_off.name: (
        EVENT_TYPE.provider_drop_off.name,
    ),
    # There is no equivalent on the agency side of maintenance_drop_off.
    # This is the closest
    PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name: (
        EVENT_TYPE.provider_drop_off.name,
        EVENT_TYPE_REASON.maintenance.name,
    ),
    PROVIDER_EVENT_TYPE_REASON.agency_drop_off.name: (
        PROVIDER_EVENT_TYPE_REASON.agency_drop_off.name,
    ),
    # reserved #
    PROVIDER_EVENT_TYPE_REASON.user_pick_up.name: (EVENT_TYPE.trip_start.name,),
    # unavailable #
    PROVIDER_EVENT_TYPE_REASON.maintenance.name: (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.maintenance.name,
    ),
    PROVIDER_EVENT_TYPE_REASON.low_battery.name: (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.low_battery.name,
    ),
    # removed #
    PROVIDER_EVENT_TYPE_REASON.service_end.name: (EVENT_TYPE.provider_pick_up.name,),
    # There is no equivalent on the agency side of rebalance_pick_up.
    PROVIDER_EVENT_TYPE_REASON.rebalance_pick_up.name: (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.rebalance.name,
    ),
    PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name: (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.maintenance.name,
    ),
    PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name: (EVENT_TYPE.city_pick_up.name,),
}

OLD_PROVIDER_REASON_TO_AGENCY_EVENT = {
    "service_start": EVENT_TYPE.service_start.name,
    "user_drop_off": EVENT_TYPE.trip_end.name,
    "rebalance_drop_off": PROVIDER_EVENT_TYPE_REASON.rebalance_drop_off.name,
    "maintenance_drop_off": PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name,
    "user_pick_up": EVENT_TYPE.trip_start.name,
    "maintenance": PROVIDER_EVENT_TYPE_REASON.maintenance.name,
    "low_battery": PROVIDER_EVENT_TYPE_REASON.low_battery.name,
    "service_end": EVENT_TYPE.service_end.name,
    "rebalance_pick_up": PROVIDER_EVENT_TYPE_REASON.rebalance_pick_up.name,
    "maintenance_pick_up": PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name,
    "agency_pick_up": PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name,
}


# Converts an Agency event to a Provider event_type_reason
# Try to stay consistent with
# https://github.com/CityOfLosAngeles/mds-core/blob/master/packages/mds-provider/utils.ts
AGENCY_EVENT_TO_PROVIDER_REASON = {
    (EVENT_TYPE.register.name,): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (EVENT_TYPE.service_start.name,): PROVIDER_EVENT_TYPE_REASON.service_start.name,
    (EVENT_TYPE.service_end.name,): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.low_battery.name,
    ): PROVIDER_EVENT_TYPE_REASON.low_battery.name,
    (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.maintenance.name,
    ): PROVIDER_EVENT_TYPE_REASON.maintenance.name,
    (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.compliance.name,
    ): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.off_hours.name,
    ): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (
        EVENT_TYPE.provider_drop_off.name,
    ): PROVIDER_EVENT_TYPE_REASON.rebalance_drop_off.name,
    # Not in the Agency spec but in the provider spec
    (
        EVENT_TYPE.provider_drop_off.name,
        EVENT_TYPE_REASON.maintenance.name,
    ): PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name,
    (EVENT_TYPE.provider_pick_up.name,): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.rebalance.name,
    ): PROVIDER_EVENT_TYPE_REASON.rebalance_pick_up.name,
    (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.maintenance.name,
    ): PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name,
    (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.charge.name,
    ): PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name,
    (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.compliance.name,
    ): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    # no city_pick_up in provider's spec
    (EVENT_TYPE.city_pick_up.name,): PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name,
    (EVENT_TYPE.reserve.name,): PROVIDER_EVENT_TYPE_REASON.user_pick_up.name,
    (
        EVENT_TYPE.cancel_reservation.name,
    ): PROVIDER_EVENT_TYPE_REASON.user_drop_off.name,
    (EVENT_TYPE.trip_start.name,): PROVIDER_EVENT_TYPE_REASON.user_pick_up.name,
    (EVENT_TYPE.trip_enter.name,): PROVIDER_EVENT_TYPE_REASON.user_pick_up.name,
    (EVENT_TYPE.trip_leave.name,): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (EVENT_TYPE.trip_end.name,): PROVIDER_EVENT_TYPE_REASON.user_drop_off.name,
    (EVENT_TYPE.deregister.name,): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (
        EVENT_TYPE.deregister.name,
        EVENT_TYPE_REASON.missing.name,
    ): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    (
        EVENT_TYPE.deregister.name,
        EVENT_TYPE_REASON.decommissioned.name,
    ): PROVIDER_EVENT_TYPE_REASON.service_end.name,
    # Not in the MDS agency spec: probably not used
    (
        PROVIDER_EVENT_TYPE_REASON.agency_drop_off.name,
    ): PROVIDER_EVENT_TYPE_REASON.agency_drop_off.name,  # not used
    (
        PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name,
    ): PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name,  # not used
    (
        EVENT_TYPE.battery_charged.name,
    ): PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name,  # not used
}

OLD_AGENCY_EVENT_TO_PROVIDER_REASON = {
    EVENT_TYPE.service_start.name: "service_start",
    EVENT_TYPE.cancel_reservation.name: "user_drop_off",
    EVENT_TYPE.trip_end.name: "user_drop_off",
    PROVIDER_EVENT_TYPE_REASON.rebalance_drop_off.name: "rebalance_drop_off",
    PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name: "maintenance_drop_off",
    EVENT_TYPE.battery_charged.name: "maintenance_drop_off",
    EVENT_TYPE.reserve.name: "user_pick_up",
    EVENT_TYPE.trip_start.name: "user_pick_up",
    EVENT_TYPE.trip_enter.name: "user_pick_up",
    EVENT_TYPE.trip_leave.name: "user_pick_up",  # This is on really bad...
    PROVIDER_EVENT_TYPE_REASON.low_battery.name: "low_battery",
    PROVIDER_EVENT_TYPE_REASON.maintenance.name: "maintenance",
    EVENT_TYPE.deregister.name: "service_end",
    EVENT_TYPE.service_end.name: "service_end",
    EVENT_TYPE.register.name: "service_end",
    PROVIDER_EVENT_TYPE_REASON.rebalance_pick_up.name: "rebalance_pick_up",
    PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name: "maintenance_pick_up",
    PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name: "agency_pick_up",
}

# Inside the Provider API, maps the event_type_reason to the corresponding event_type
# https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/provider
PROVIDER_EVENT_TYPE_REASON_TO_EVENT_TYPE = {
    # available #
    PROVIDER_EVENT_TYPE_REASON.service_start.name: DEVICE_STATUS.available.name,
    PROVIDER_EVENT_TYPE_REASON.user_drop_off.name: DEVICE_STATUS.available.name,
    PROVIDER_EVENT_TYPE_REASON.rebalance_drop_off.name: DEVICE_STATUS.available.name,
    PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name: DEVICE_STATUS.available.name,
    PROVIDER_EVENT_TYPE_REASON.agency_drop_off.name: DEVICE_STATUS.available.name,
    # reserved #
    PROVIDER_EVENT_TYPE_REASON.user_pick_up.name: DEVICE_STATUS.reserved.name,
    # unavailable #
    PROVIDER_EVENT_TYPE_REASON.maintenance.name: DEVICE_STATUS.unavailable.name,
    PROVIDER_EVENT_TYPE_REASON.low_battery.name: DEVICE_STATUS.unavailable.name,
    # removed #
    PROVIDER_EVENT_TYPE_REASON.service_end.name: DEVICE_STATUS.removed.name,
    PROVIDER_EVENT_TYPE_REASON.rebalance_pick_up.name: DEVICE_STATUS.removed.name,
    PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name: DEVICE_STATUS.removed.name,
    PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name: DEVICE_STATUS.removed.name,
    # others #
    PROVIDER_EVENT_TYPE_REASON.telemetry.name: DEVICE_STATUS.unknown.name,
    PROVIDER_EVENT_TYPE_REASON.battery_charged.name: DEVICE_STATUS.available.name,
    # old agency api
    EVENT_TYPE.provider_drop_off.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.provider_pick_up.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.register.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.trip_leave.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.deregister.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.reserve.name: DEVICE_STATUS.reserved.name,
    EVENT_TYPE.trip_enter.name: DEVICE_STATUS.reserved.name,
    EVENT_TYPE.cancel_reservation.name: DEVICE_STATUS.available.name,
}


OLD_TO_NEW_AGENCY_EVENT = {
    (EVENT_TYPE.service_start.name,): (EVENT_TYPE.service_start.name,),
    (EVENT_TYPE.trip_end.name,): (EVENT_TYPE.trip_end.name,),
    (PROVIDER_EVENT_TYPE_REASON.rebalance_drop_off.name,): (
        EVENT_TYPE.provider_drop_off.name,
    ),
    (PROVIDER_EVENT_TYPE_REASON.maintenance_drop_off.name,): (
        EVENT_TYPE.provider_drop_off.name,
        EVENT_TYPE_REASON.maintenance.name,
    ),
    (EVENT_TYPE.trip_start.name,): (EVENT_TYPE.trip_start.name,),
    (PROVIDER_EVENT_TYPE_REASON.maintenance.name,): (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.maintenance.name,
    ),
    (PROVIDER_EVENT_TYPE_REASON.low_battery.name,): (
        EVENT_TYPE.service_end.name,
        EVENT_TYPE_REASON.low_battery.name,
    ),
    (EVENT_TYPE.service_end.name,): (EVENT_TYPE.provider_pick_up.name,),
    (PROVIDER_EVENT_TYPE_REASON.rebalance_pick_up.name,): (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.rebalance.name,
    ),
    (PROVIDER_EVENT_TYPE_REASON.maintenance_pick_up.name,): (
        EVENT_TYPE.provider_pick_up.name,
        EVENT_TYPE_REASON.maintenance.name,
    ),
    (PROVIDER_EVENT_TYPE_REASON.agency_pick_up.name,): (EVENT_TYPE.city_pick_up.name,),
    (EVENT_TYPE.trip_enter.name,): (EVENT_TYPE.trip_enter.name,),
    (EVENT_TYPE.cancel_reservation.name,): (EVENT_TYPE.cancel_reservation.name,),
    (EVENT_TYPE.telemetry.name,): (EVENT_TYPE.telemetry.name,),
    (EVENT_TYPE.reserve.name,): (EVENT_TYPE.reserve.name,),
    (EVENT_TYPE.deregister.name,): (EVENT_TYPE.deregister.name,),
    (EVENT_TYPE.battery_charged.name,): (EVENT_TYPE.battery_charged.name,),
    (EVENT_TYPE.trip_leave.name,): (EVENT_TYPE.trip_leave.name,),
    (EVENT_TYPE.register.name,): (EVENT_TYPE.register.name,),
}


def get_new_event_from_old(old_event):
    """
    Maps an old agency event to the new agency event.
    """
    if isinstance(old_event, str):
        old_event = (old_event,)
    return OLD_TO_NEW_AGENCY_EVENT.get(old_event, old_event)


def get_old_event_from_new(event):
    """
    Maps an new agency event to the old one.
    """
    old_event = (
        [
            old_evt
            for old_evt, new_evt in OLD_TO_NEW_AGENCY_EVENT.items()
            if new_evt == event
        ]
        + [event]  # if the event is not in the mapping
    )[0]

    return old_event


def get_provider_reason_from_both_mappings(event_record):
    """
    Takes an EventRecord as input (stored with the Agency nomenclature),
    and returns the provider event_type_reason.
    Supports both the old and the new mapping.
    """
    use_old_mapping = True
    try:
        event = (event_record.event_type,)
    except AttributeError:
        raise ValueError("Not a valid EventRecord object")

    try:
        if isinstance(event_record.event_type_reason, str):
            event = (event_record.event_type, event_record.event_type_reason)
            use_old_mapping = False
        else:
            if (event_record.event_type,) in AGENCY_EVENT_TO_PROVIDER_REASON.keys():
                use_old_mapping = False
    except AttributeError:
        pass

    if use_old_mapping:
        return OLD_AGENCY_EVENT_TO_PROVIDER_REASON[event_record.event_type]
    else:
        return AGENCY_EVENT_TO_PROVIDER_REASON[event]


def get_same_mapping_event():
    """
    Returns the list of event type reasons that do are not supposed to change
    during the migration.
    """
    return [e for e in OLD_TO_NEW_AGENCY_EVENT.keys() if e == get_new_event_from_old(e)]



# Generated by Django 2.1.5 on 2019-01-22 13:00

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0004_eventrecord_point'),
    ]

    operations = [
        migrations.AlterField(
            model_name='device',
            name='dn_status',
            field=mds.models.UnboundedCharField(choices=[('available', 'Available'), ('reserved', 'Reserved'), ('unavailable', 'Unavailable'), ('removed', 'Removed'), ('unknown', 'Unknown')], default='unavailable'),
        ),
    ]



import django_filters.rest_framework
from rest_framework import pagination, serializers, viewsets

from mds.access_control import scopes
from mds.access_control.permissions import require_scopes
import mds.models


class EventRecordSerializer(serializers.ModelSerializer):
    class Meta:
        model = mds.models.EventRecord
        fields = ["timestamp", "event_type"]


class EventRecordViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = EventRecordSerializer
    queryset = mds.models.EventRecord.objects.order_by("-timestamp")
    pagination_class = pagination.LimitOffsetPagination
    permission_classes = [require_scopes(scopes.SCOPE_PRV_API)]
    filter_backends = (django_filters.rest_framework.DjangoFilterBackend,)
    filterset_fields = ("device",)



# Generated by Django 2.1.10 on 2019-07-23 17:35

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("mds", "0036_pre_save_provider_cursors")]

    operations = [
        migrations.RemoveField(model_name="provider", name="last_start_time_polled")
    ]




from typing import List, Dict

import jwt

from .auth_means import BaseAuthMean, PublicKeyJwtBaseAuthMean, SecretKeyJwtBaseAuthMean


def jwt_multi_decode(auth_means: List[BaseAuthMean], encoded_jwt: str) -> (Dict, str):
    """
    Try all available secret and public keys to decode the JWT
    """

    assert auth_means, "At least one JWT key must be provided"

    exception_holder = jwt.InvalidSignatureError()

    header = jwt.get_unverified_header(encoded_jwt)
    alg = header["alg"]

    for auth_mean in auth_means:
        if alg == "RS256" and isinstance(auth_mean, PublicKeyJwtBaseAuthMean):
            try:
                return (
                    jwt.decode(encoded_jwt, auth_mean.public_key, algorithms="RS256"),
                    auth_mean.introspect_url,
                )
            except jwt.InvalidSignatureError as e:
                exception_holder = e
        elif isinstance(auth_mean, SecretKeyJwtBaseAuthMean):
            try:
                return (
                    jwt.decode(encoded_jwt, auth_mean.secret_key, algorithms="HS256"),
                    auth_mean.introspect_url,
                )
            except jwt.InvalidSignatureError as e:
                exception_holder = e

    raise exception_holder



# Generated by Django 2.2.2 on 2019-06-07 14:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("mds", "0026_alter_polygon_to_multipolygon")]

    operations = [
        migrations.AddField(
            model_name="eventrecord",
            name="first_saved_at",
            field=models.DateTimeField(blank=True, null=True),
        )
    ]



from rest_framework import decorators, pagination, response, serializers, viewsets

from mds import enums, models, utils
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_PRV_API
from mds.apis import utils as apis_utils
from mds.provider_mapping import (
    AGENCY_EVENT_TO_PROVIDER_REASON,
    PROVIDER_EVENT_TYPE_REASON_TO_EVENT_TYPE,
    OLD_AGENCY_EVENT_TO_PROVIDER_REASON,
    get_provider_reason_from_both_mappings,
)


class DeviceStatusChangesSerializer(serializers.ModelSerializer):
    id = serializers.CharField()
    recorded = apis_utils.UnixTimestampMilliseconds(source="saved_at")
    first_recorded = apis_utils.UnixTimestampMilliseconds(source="first_saved_at")
    associated_trip = serializers.SerializerMethodField()
    device_id = serializers.CharField(source="device.id")
    event_location = serializers.SerializerMethodField()
    event_time = apis_utils.UnixTimestampMilliseconds(source="timestamp")
    event_type = serializers.SerializerMethodField()
    event_type_reason = serializers.SerializerMethodField()
    propulsion_type = serializers.ListSerializer(
        source="device.propulsion",
        child=serializers.ChoiceField(choices=enums.choices(enums.DEVICE_PROPULSION)),
    )
    provider_id = serializers.CharField(source="device.provider.id")
    provider_name = serializers.CharField(source="device.provider.name")
    vehicle_id = serializers.CharField(source="device.identification_number")
    vehicle_type = serializers.CharField(source="device.category")

    class Meta:
        model = models.EventRecord
        fields = (
            "id",
            "recorded",
            "first_recorded",
            "associated_trip",
            "device_id",
            "event_location",
            "event_time",
            "event_type",
            "event_type_reason",
            "propulsion_type",
            "provider_name",
            "provider_id",
            "vehicle_id",
            "vehicle_type",
        )

    def create(self, data):
        raise NotImplementedError()

    def update(self, instance, data):
        raise NotImplementedError()

    def get_associated_trip(self, obj):
        return obj.properties.get("trip_id", None)

    def get_event_location(self, obj):
        telemetry = obj.properties.get("telemetry", {})
        return {
            "type": "Feature",
            "properties": {"timestamp": telemetry.get("timestamp")},
            "geometry": obj.point_as_geojson,
        }

    def get_event_type(self, obj):
        reason = self.get_event_type_reason(obj)
        return PROVIDER_EVENT_TYPE_REASON_TO_EVENT_TYPE[reason]

    def get_event_type_reason(self, obj):
        return get_provider_reason_from_both_mappings(obj)


class CustomPagination(pagination.PageNumberPagination):
    page_size = 5000
    page_size_query_param = "take"
    max_page_size = 10000

    def get_paginated_response(self, data):
        return response.Response(
            {
                "version": "0.3.0",
                "links": {
                    "next": self.get_next_link(),
                    "prev": self.get_previous_link(),
                },
                "data": {"status_changes": data},
            }
        )


class ProviderApiViewSet(viewsets.ViewSet):
    permission_classes = [require_scopes(SCOPE_PRV_API)]

    @decorators.action(detail=False, methods=["get"])
    def status_changes(self, request, *args, **kwargs):
        skip = request.query_params.get("skip")
        start_recorded = request.query_params.get("start_recorded")
        start_time = request.query_params.get("start_time")
        end_time = request.query_params.get("end_time")

        events = models.EventRecord.objects.select_related("device__provider").filter(
            # Only forward events that can be polled from a "provider API"
            # We don't want to filter out the old or the new events though.
            event_type__in=[x[0] for x in AGENCY_EVENT_TO_PROVIDER_REASON.keys()]
            + list(OLD_AGENCY_EVENT_TO_PROVIDER_REASON.keys())
        )

        # We support either recorded, time search or offset but not at the same time
        if skip:
            order_by = "id"
            events = events.filter(id__gt=int(skip))
        elif start_recorded:
            order_by = "saved_at"
            start_recorded = utils.from_mds_timestamp(int(start_recorded))
            events = events.filter(saved_at__gte=start_recorded)
        else:
            order_by = "timestamp"
            if start_time:
                start_time = utils.from_mds_timestamp(int(start_time))
                events = events.filter(timestamp__gte=start_time)
            if end_time:
                end_time = utils.from_mds_timestamp(int(end_time))
                events = events.filter(timestamp__lte=end_time)

        paginator = CustomPagination()
        page = paginator.paginate_queryset(events.order_by(order_by), request)
        data = DeviceStatusChangesSerializer(page, many=True).data
        return paginator.get_paginated_response(data)



"""
Pulling data for registered providers

This is the opposite of provider pushing their data to the agency API.
"""
import logging

from django.core import management

from mds import models
from mds.provider_poller import poller


logger = logging.getLogger(__name__)


class Command(management.BaseCommand):
    help = "Poll providers to fetch their latest data."

    def add_arguments(self, parser):
        parser.add_argument(
            "--raise-on-error",
            action="store_true",
            help=("Raise exceptions instead of ignoring them."),
        )

    def handle(self, *args, **options):
        for provider in models.Provider.objects.all():
            logger.debug("Polling provider %s... ", provider.name)
            try:
                poller.StatusChangesPoller(provider).poll()
            except Exception:  # pylint: disable=broad-except
                # In dev, test... environments, we want explicit errors
                if options["raise_on_error"]:
                    raise
                # But in production, we just log and try the next one
                logger.exception("Error in polling provider %s", provider.name)
            else:
                logger.debug("Polling provider %s succeeded.", provider.name)



# Generated by Django 2.1.3 on 2019-04-26 08:07
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [("mds", "0020_fields_consistency")]

    operations = [
        migrations.AlterField(
            model_name="eventrecord",
            name="source",
            field=mds.models.UnboundedCharField(
                choices=[
                    ("agency_api", "Agency API"),
                    ("provider_api", "Provider API"),
                ],
                default="agency_api",
            ),
        ),
        migrations.RunSQL(
            """
            UPDATE mds_eventrecord SET source='agency_api' WHERE source='push';
            UPDATE mds_eventrecord SET source='provider_api' WHERE source='pull';
            """,
            reverse_sql="""
            UPDATE mds_eventrecord SET source='push' WHERE source='agency_api';
            UPDATE mds_eventrecord SET source='pull' WHERE source='provider_api';
        """,
        ),
    ]
